import unittest
from unittest import mock
import tempfile
from os.path import isfile
from os import remove
import numpy as np
from geviewer import main, geviewer


class TestGeViewerMain(unittest.TestCase):

    def setUp(self):
        """Create a GeViewer object.
        """
        self.gev = geviewer.GeViewer(['tests/sample.wrl'], off_screen=True)


    def test_print_instructions(self):
        """Test the print_instructions function.
        """
        main.print_instructions()


    def test_key_inputs(self):
        """Test the key inputs for the GeViewer object.
        """
        for i in range(2):
            with self.subTest():
                track_status = self.gev.visible[0]
                self.gev.toggle_tracks()
                track_actors = self.gev.actors[0]
                self.assertTrue(track_actors.visibility!=track_status)
            with self.subTest():
                step_status = self.gev.visible[1]
                self.gev.toggle_step_markers()
                step_actors = self.gev.actors[1]
                self.assertTrue(step_actors.visibility!=step_status)
            with self.subTest():
                colors = ['lightskyblue','white']
                bkg_status = self.gev.bkg_on
                self.gev.toggle_background()
                self.assertEqual(self.gev.plotter.background_color, colors[bkg_status])


    @mock.patch('geviewer.utils.prompt_for_screenshot_path')
    def test_save_screenshot(self, mocked_input):
        """Test the save_graphic method with mocked file name inputs.
        """
        with tempfile.TemporaryDirectory() as temp:
            file_names = [temp + '/file.png',\
                          temp + '/file.svg',\
                          temp + '/file.eps',\
                          temp + '/file.ps',\
                          temp + '/file.pdf',\
                          temp + '/file.tex']
            mocked_input.side_effect = file_names
            for i in file_names:
                self.gev.save_screenshot()
                with self.subTest():
                    self.assertTrue(isfile(i))


    @mock.patch('geviewer.utils.prompt_for_window_size')
    def test_set_window_size(self, mocked_input):
        """Test the set_window_size method with a mocked window size input.
        """
        mocked_input.return_value = [800,600]
        self.gev.set_window_size()
        self.assertEqual(self.gev.plotter.window_size,[800,600])
    

    @mock.patch('geviewer.utils.prompt_for_camera_view')
    def test_set_camera_view(self, mocked_input):
        """Test the set_camera_view method with a mocked camera viewpoint input.
        """
        pos = (1,2,3)
        up_input = (4,5,6)
        focus = (7,8,9)
        up_output = tuple(np.array(up_input)/np.linalg.norm(up_input))
        mocked_input.return_value = [pos,up_input,focus]                                   
        with self.subTest():
            self.gev.set_camera_view()
            self.assertEqual(self.gev.plotter.camera.position,pos)
            self.assertEqual(self.gev.plotter.camera.up,up_output)
            self.assertEqual(self.gev.plotter.camera.focal_point,focus)


    @mock.patch('geviewer.utils.prompt_for_html_path')
    def test_export_to_html(self, mocked_input):
        """Test the export_to_html method with a mocked file path input.
        """
        with tempfile.TemporaryDirectory() as temp:
            mocked_input.return_value = temp + '/file.html'
            self.gev.export_to_html()
            self.assertTrue(isfile(mocked_input.return_value))


    def test_save_and_load(self):
        """Test the save session method.
        """
        with tempfile.TemporaryDirectory() as temp:
            with self.subTest():
                self.gev.save(temp + '/file.gev')
                self.assertTrue(isfile(temp + '/file.gev'))
            with self.subTest():
                gev = geviewer.GeViewer([temp + '/file.gev'], off_screen=True)
                self.assertTrue(all([gev.actors[i].visibility for i in range(2)]))
    

if __name__ == '__main__':
    unittest.main()


from geviewer.geviewer import GeViewer
import argparse


def print_instructions():
    """Prints the banner and instructions for the user.
    """
    print()
    print('###################################################')
    print('#    _____   __      ___                          #')
    print('#   / ____|  \\ \\    / (_)                         #')
    print('#  | |  __  __\\ \\  / / _  _____      _____ _ __   #')
    print('#  | | |_ |/ _ \\ \\/ / | |/ _ \\ \\ /\\ / / _ \\  __|  #')
    print('#  | |__| |  __/\\  /  | |  __/\\ V  V /  __/ |     #')
    print('#   \\_____|\\___| \\/   |_|\\___| \\_/\\_/ \\___|_|     #')
    print('#                                                 #')
    print('###################################################')
    print()
    print('Instructions:')
    print('-------------')
    print('* Click and drag to rotate the view, shift + click')
    print('  and drag to pan,  ctrl + click and drag to roll,')
    print('  and scroll to zoom')
    print('* Press "c" to capture a screenshot of the current view')
    print('* Press "t" to toggle the trajectories on or off')
    print('* Press "m" to toggle the step markers on or off')
    print('* Press "b" to toggle the background on or off')
    print('* Press "w" to toggle between wireframe and solid modes')
    print('* Press "v" to switch to an isometric view')
    print('* Press "d" to set the display window size')
    print('* Press "i" to set the camera viewpoint')
    print('* Press "p" to print the current display settings')
    print('* Press "h" to export the viewer to an HTML file')
    print('* Press "q" to quit the viewer')
    print()


def main():
    """Command line interface for GeViewer.

    This function sets up the command-line interface for running the GeViewer application.
    It uses the argparse module to parse command-line arguments, allowing users to specify
    the files to be displayed and various options for running the viewer. 

    Command-line options include:
    - `filenames`: The file or list of files to be displayed. This argument is required.
    - `-d` or `--destination`: Optional argument specifying the path to save the session.
    If provided without a value, the default filename is 'viewer.gev'.
    - `-o` or `--off-screen`: Optional flag to run the viewer in offscreen mode.
    - `-s` or `--safe-mode`: Optional flag to enable a more robust VRML parsing mode at
    the expense of some interactive features.
    - `-w` or `--no-warnings`: Optional flag to suppress warnings and prevent the program
    from pausing to display them.
    """
    print_instructions()

    parser = argparse.ArgumentParser(description='View Geant4 simulation results.')
    parser.add_argument('filenames', nargs='+', help='the file or list of files to be displayed')
    parser.add_argument('-d', '--destination', nargs='?', help='save the session to this location', \
                        default=None, const='viewer.gev')
    parser.add_argument('-o', '--off-screen', help='run in offscreen mode.', action='store_true')
    parser.add_argument('-s', '--safe-mode', help='use more robust VRML parsing ' \
                        + 'at the expense of some interactive features', action='store_true')
    parser.add_argument('-w', '--no-warnings', help='do not pause the program to display warnings', \
                        action='store_true')
    args = parser.parse_args()
    filenames = args.filenames
    destination = args.destination
    off_screen = args.off_screen
    safe_mode = args.safe_mode
    no_warnings = args.no_warnings

    GeViewer(filenames, destination=destination, off_screen=off_screen,\
             safe_mode=safe_mode, no_warnings=no_warnings)


if __name__ == '__main__':
    main()



import numpy as np
import pyvista as pv
from tqdm import tqdm
import re


def create_meshes(polyline_blocks, marker_blocks, solid_blocks):
    """Creates and returns meshes for polylines, markers, and solids.

    This function processes blocks of data for polylines, markers, and solids,
    building corresponding meshes for each.

    :param polyline_blocks: List of blocks containing polyline data.
    :type polyline_blocks: list
    :param marker_blocks: List of blocks containing marker data.
    :type marker_blocks: list
    :param solid_blocks: List of blocks containing solid data.
    :type solid_blocks: list

    :returns: A tuple containing three elements:
        - polyline_mesh (:class:`pyvista.PolyData`): Mesh for polylines.
        - marker_mesh (:class:`pyvista.UnstructuredGrid`): Mesh for markers.
        - solid_mesh (:class:`pyvista.PolyData`): Mesh for solids.
    :rtype: tuple
    """
    total = len(polyline_blocks) + len(marker_blocks) + len(solid_blocks)
    with tqdm(total=total, desc='Building meshes...') as pbar:
        polyline_mesh = build_mesh(polyline_blocks, 'polyline', pbar)
        marker_mesh = build_markers(marker_blocks, pbar)
        solid_mesh = build_mesh(solid_blocks, 'solid', pbar)

    return polyline_mesh, marker_mesh, solid_mesh


def combine_mesh_arrays(points, cells, colors, pbar=None):
    """Combines multiple mesh arrays into a single mesh.

    This function takes lists of points, indices of faces or line segments
    (called cells), and colors, and combines them into a single set of points,
    cells, and colors, adjusting indices appropriately.

    :param points: A list of arrays containing point coordinates.
    :type points: list of numpy.ndarray
    :param cells: A list of lists containing cell indices.
    :type cells: list of list
    :param colors: A list of arrays containing color data.
    :type colors: list of numpy.ndarray
    :param pbar: (Optional) A tqdm progress bar instance.
    :type pbar: tqdm.tqdm, optional

    :returns: A tuple containing three elements:
        - points (numpy.ndarray): Combined array of point coordinates.
        - cells (numpy.ndarray): Combined array of cell indices.
        - colors (numpy.ndarray): Combined array of color data.
    :rtype: tuple
    """
    offsets = np.cumsum([0] + [len(p) for p in points[:-1]]).astype(int)
    points = np.concatenate(points)
    for i, cell in enumerate(cells):
        j = 0
        while j < len(cell):
            k = cell[j]
            cell[j + 1:j + k + 1] = (np.array(cell[j + 1:j + k + 1]) + offsets[i]).tolist()
            j += k + 1
    cells = np.concatenate(cells).astype(int)
    colors = np.concatenate(colors)
    if pbar:
        pbar.update(1)
    return points, cells, colors


def build_mesh(blocks, which, pbar=None):
    """Builds a mesh from given blocks of data.

    This function processes blocks of data, creating a mesh based on the specified
    type ('polyline' or 'solid').

    :param blocks: List of blocks containing data for the mesh.
    :type blocks: list
    :param which: Type of mesh to build ('polyline' or 'solid').
    :type which: str
    :param pbar: (Optional) A tqdm progress bar instance.
    :type pbar: tqdm.tqdm, optional

    :returns: The created mesh with combined points, cells, and colors.
    :rtype: pyvista.PolyData
    """
    points = [None for i in range(len(blocks))]
    cells = [None for i in range(len(blocks))]
    colors = [None for i in range(len(blocks))]

    if which == 'polyline':
        func = process_polyline_block
    elif which == 'solid':
        func = process_solid_block

    for i, block in enumerate(blocks):
        points[i], cells[i], color = func(block)
        colors[i] = [color]*len(points[i])
        if pbar:
            pbar.update(1)

    if len(points) == 0:
        return None
    
    points, cells, colors = combine_mesh_arrays(points, cells, colors)
    if func==process_polyline_block:
        mesh = pv.PolyData(points, lines=cells)
    elif func==process_solid_block:
        mesh = pv.PolyData(points, faces=cells)
    mesh.point_data.set_scalars(colors, name='color')

    return mesh


def build_markers(blocks, pbar=None):
    """Builds a mesh for markers from given blocks of data.

    This function processes blocks of marker data, creating a mesh of spheres for each marker.

    :param blocks: List of blocks containing marker data.
    :type blocks: list
    :param pbar: (Optional) A tqdm progress bar instance.
    :type pbar: tqdm.tqdm, optional

    :returns: The created mesh with combined centers, radii, and colors.
    :rtype: pyvista.UnstructuredGrid
    """
    centers = [None for i in range(len(blocks))]
    radii = [None for i in range(len(blocks))]
    colors = [None for i in range(len(blocks))]
    
    for i, block in enumerate(blocks):
        centers[i], radii[i], colors[i] = process_marker_block(block)
    if len(centers) == 0:
        return None
    
    mesh = pv.MultiBlock()
    for i in range(len(centers)):
        mesh.append(pv.Sphere(radius=radii[i], center=centers[i]))
        colors[i] = [colors[i]]*mesh[-1].n_points
        if pbar:
            pbar.update(1)

    colors = np.concatenate(colors)
    mesh = mesh.combine()
    mesh.point_data.set_scalars(colors, name='color')

    return mesh


def extract_blocks(file_content):
    """Extracts polyline, marker, and solid blocks from the given file content.

    This function processes the provided file content, which is expected to
    be in a text format, and extracts blocks of different types based on
    specific keywords. It separates the blocks into categories: polyline,
    marker, and solid blocks, and also identifies the viewpoint block.

    :param file_content: The content of the file as a single string.
    :type file_content: str
    :return: A tuple containing four elements:
        - The viewpoint block (if found) as a string or `None` if not found.
        - A list of polyline blocks as strings.
        - A list of marker blocks as strings.
        - A list of solid blocks as strings.
    :rtype: tuple
    """
    polyline_blocks = []
    marker_blocks = []
    solid_blocks = []
    viewpoint_block = None

    lines = file_content.split('\n')
    block = []
    inside_block = False
    brace_count = 0

    for line in tqdm(lines, desc='Parsing data......'):
        stripped_line = line.strip()

        if stripped_line.startswith('Shape') or stripped_line.startswith('Anchor')\
            or stripped_line.startswith('Viewpoint'):
            inside_block = True
            brace_count = 0
        
        if inside_block:
            block.append(line)
            brace_count += line.count('{') - line.count('}')
            
            if brace_count == 0:
                block_content = '\n'.join(block)
                
                if 'IndexedLineSet' in block_content:
                    polyline_blocks.append(block_content)
                elif 'Sphere' in block_content:
                    marker_blocks.append(block_content)
                elif 'IndexedFaceSet' in block_content:
                    solid_blocks.append(block_content)
                elif 'Viewpoint' in block_content:
                    viewpoint_block = block_content

                block = []
                inside_block = False

    return viewpoint_block, polyline_blocks, marker_blocks, solid_blocks


def process_polyline_block(block):
    """Processes a polyline block to create a polyline mesh.

    This function takes a block of polyline data and converts it into a
    PyVista`PolyData` object representing the polyline mesh. It also
    extracts the color information associated with the mesh.

    :param block: The polyline block content as a string.
    :type block: str
    :return: A tuple containing:
        - A `pv.PolyData` object representing the polyline mesh.
        - The color associated with the polyline mesh as a list or array.
    :rtype: tuple
    """
    points, indices, color = parse_polyline_block(block)
    lines = []
    for i in range(len(indices) - 1):
        if indices[i] != -1 and indices[i + 1] != -1:
            lines.extend([2, indices[i], indices[i + 1]])
    
    return points, lines, color


def process_marker_block(block):
    """Processes a marker block to create a marker mesh.

    This function takes a block of marker data and creates a spherical
    marker mesh using PyVista. It also extracts the color information
    associated with the marker.

    :param block: The marker block content as a string.
    :type block: str
    :return: A tuple containing:
        - A `pv.Sphere` object representing the marker mesh.
        - The color associated with the marker mesh as a list or array.
    :rtype: tuple
    """
    center, radius, color = parse_marker_block(block)

    return center, radius, color


def process_solid_block(block):
    """Processes a solid block to create a solid mesh.

    This function takes a block of solid data and creates a mesh for a
    solid object using PyVista. It also extracts the color information
    associated with the solid.

    :param block: The solid block content as a string.
    :type block: str
    :return: A tuple containing:
        - A `pv.PolyData` object representing the solid mesh.
        - The color associated with the solid mesh as a list or array.
    :rtype: tuple
    """
    points, indices, color = parse_solid_block(block)
    faces = []
    current_face = []
    for index in indices:
        if index == -1:
            if len(current_face) == 3:
                faces.extend([3] + current_face)
            elif len(current_face) == 4:
                faces.extend([4] + current_face)
            current_face = []
        else:
            current_face.append(index)
    faces = np.array(faces)

    return points, faces, color


def parse_viewpoint_block(block):
    """Parses the viewpoint block to extract the field of view, position,
    and orientation.

    This function extracts the field of view (FOV), position, and orientation
    from a given viewpoint block in a 3D scene description. The FOV is converted
    from radians to degrees.

    :param block: The viewpoint block content as a string.
    :type block: str
    :return: A tuple containing:
        - The field of view in degrees as a float (or None if not found).
        - The position as a list of three floats [x, y, z] (or None if not found).
        - The orientation as a list of four floats [x, y, z, angle] in radians
        (or None if not found).
    :rtype: tuple
    """
    fov = None
    position = None
    orientation = None

    if block is not None:
        fov_match = re.search(r'fieldOfView\s+([\d.]+)', block)
        if fov_match:
            fov = float(fov_match.group(1))*180/np.pi
        
        position_match = re.search(r'position\s+([\d.-]+)\s+([\d.-]+)\s+([\d.-]+)', block)
        if position_match:
            position = [float(position_match.group(1)), float(position_match.group(2)), \
                        float(position_match.group(3))]

        orientation_match = re.search(r'orientation\s+([\d.-]+)\s+([\d.-]+)\s+([\d.-]+)\s+([\d.-]+)', block)
        if orientation_match:
            orientation = [float(orientation_match.group(1)), float(orientation_match.group(2)), \
                           float(orientation_match.group(3)), float(orientation_match.group(4))]
    
    return fov, position, orientation


def parse_polyline_block(block):
    """Parses a polyline block to extract particle track information, including
    coordinates, indices, and color.

    This function processes a block of text representing a polyline in a 3D
    scene description. It extracts the coordinates of the points that define
    the polyline, the indices that describe the lines between these points, 
    and the color associated with the polyline.

    :param block: The polyline block content as a string.
    :type block: str
    :return: A tuple containing:
        - `coords`: An array of shape (N, 3) representing the coordinates of the
        polyline points.
        - `indices`: An array of integers representing the indices that define
        the polyline segments.
        - `color`: An array of four floats representing the RGBA color of the
        polyline, where the alpha is set to 1.
    :rtype: tuple
    """
    coords = []
    coord_inds = []
    color = [1, 1, 1]

    lines = block.split('\n')
    reading_points = False
    reading_indices = False

    for line in lines:
        line = line.strip()
        if line.startswith('point ['):
            reading_points = True
            continue
        elif line.startswith(']'):
            reading_points = False
            reading_indices = False
            continue
        elif line.startswith('coordIndex ['):
            reading_indices = True
            continue
        elif 'diffuseColor' in line:
            color = list(map(float, re.findall(r'[-+]?\d*\.?\d+', line)))
        if reading_points:
            point = line.replace(',', '').split()
            if len(point) == 3:
                coords.append(list(map(float, point)))
        elif reading_indices:
            indices = line.replace(',', '').split()
            coord_inds.extend(list(map(int, indices)))

    color.append(1.)

    return np.array(coords), np.array(coord_inds), np.array(color)


def parse_marker_block(block):
    """Parses a marker block to extract the position, radius, and color of a marker.

    This function processes a block of text representing a marker in a 3D scene
    description. It extracts the position of the marker, the radius of the marker
    (typically a sphere), and the color of the marker. It also accounts for
    transparency and adjusts the alpha value of the color accordingly.

    :param block: The marker block content as a string.
    :type block: str
    :return: A tuple containing:
        - `coords`: An array of shape (3,) representing the position of the marker
        in 3D space.
        - `radius`: A float representing the radius of the marker.
        - `color`: An array of four floats representing the RGBA color of the marker,
        where alpha is adjusted for transparency.
    :rtype: tuple
    """
    coords = []
    color = [1, 1, 1]
    transparency = 0
    radius = 1

    lines = block.split('\n')

    for line in lines:
        line = line.strip()
        if line.startswith('translation'):
            point = line.split()[1:]
            if len(point) == 3:
                coords = list(map(float, point))
        elif 'diffuseColor' in line:
            color = list(map(float, re.findall(r'[-+]?\d*\.?\d+', line)))
        elif 'transparency' in line:
            transparency = float(re.findall(r'[-+]?\d*\.?\d+', line)[0])
        elif 'radius' in line:
            radius = float(re.findall(r'[-+]?\d*\.?\d+', line)[0])

    color.append(1. - transparency)

    return np.array(coords), radius, np.array(color)


def parse_solid_block(block):
    """Parses a solid block to extract geometry information for a 3D
    solid object.

    This function processes a block of text representing a solid object
    in a 3D scene description. It extracts the vertex coordinates, the face
    indices that define the geometry of the solid, and the color of the solid.
    The function also handles transparency by adjusting the alpha value in the
    color array.

    :param block: The solid block content as a string.
    :type block: str
    :return: A tuple containing:
        - `coords`: An array of shape (N, 3) where N is the number of vertices,
        representing the vertex coordinates.
        - `coord_inds`: An array of shape (M,) where M is the number of indices,
        representing the indices defining the faces of the solid.
        - `color`: An array of four floats representing the RGBA color of the solid,
        where the alpha value is adjusted for transparency.
    :rtype: tuple
    """
    coords = []
    coord_inds = []
    color = [1, 1, 1]
    transparency = 0

    lines = block.split('\n')
    reading_points = False
    reading_indices = False

    for line in lines:
        line = line.strip()
        if line.startswith('point ['):
            reading_points = True
            continue
        elif line.startswith(']'):
            reading_points = False
            reading_indices = False
            continue
        elif line.startswith('coordIndex ['):
            reading_indices = True
            continue
        elif 'diffuseColor' in line:
            color = list(map(float, re.findall(r'[-+]?\d*\.?\d+', line)))
        elif 'transparency' in line:
            transparency = float(re.findall(r'[-+]?\d*\.?\d+', line)[0])
        if reading_points:
            point = line.replace(',', '').split()
            if len(point) == 3:
                coords.append(list(map(float, point)))
        elif reading_indices:
            indices = line.replace(',', '').split()
            coord_inds.extend(list(map(int, indices)))

    color.append(1. - transparency)

    return np.array(coords), np.array(coord_inds), np.array(color)


import pyvista as pv


class Plotter(pv.Plotter):
    """
    A custom plotter class that extends `pyvista.Plotter` to provide
    enhanced key event handling.

    This class inherits from `pyvista.Plotter` and modifies key event
    handling to override specific key behaviors. Keys that are not overridden
    by this class will be passed to the original key event handler.

    :param args: Positional arguments to pass to the base class constructor.
    :param kwargs: Keyword arguments to pass to the base class constructor.

    Methods:
        - `geviewer_key_event`: Handles key press events and overrides specific key behaviors.
        - `show`: Displays the plotter window with custom key event handling.

    Inherited Methods:
        - `pyvista.Plotter.show`: Displays the plotter window.
    """

    def geviewer_key_event(self, obj, event):
        """Handles key press events to override specific key behaviors.

        This method intercepts key press events and checks if the key pressed
        is in the list of overridden keys. If it is, the method performs no
        action for these keys. For other keys, the method delegates the event
        to the original key event handler.

        :param obj: The VTK object that triggered the event.
        :type obj: vtkObject
        :param event: The event name (e.g., "CharEvent").
        :type event: str
        """
        key = obj.GetInteractor().GetKeyCode()
        overriden_keys = ['c', 't', 'm', 'b', 'd', 'i', 'p', 'h',\
                          'w', 'e', '+', '-', 's', 'r', 'u', 'f']
        if key in overriden_keys:
            return
        else:
            # Commands for other keys are passed to the original key event handler
            obj.OnChar()


    def show(self, *args, **kwargs):
        """Displays the plotter window with custom key event handling.

        This method overrides the `show` method from `pyvista.Plotter` to include
        custom key event handling. It first sets up an observer for key press events
        and then calls the base class `show` method to displaythe plotter window.

        :param args: Positional arguments passed to the base class `show` method.
        :param kwargs: Keyword arguments passed to the base class `show` method.
        """
        interactor_style = self.iren.interactor.GetInteractorStyle()
        interactor_style.AddObserver("CharEvent", self.geviewer_key_event)

        # Now call the original show method
        super().show(*args, **kwargs)



import numpy as np
import pyvista as pv
import asyncio
from pathlib import Path
import os
import shutil
import zipfile
import tempfile
from geviewer import utils, parser, plotter


class GeViewer:
    """The main interface for the GeViewer application, responsible for loading,
    processing, and visualizing data files. This class manages the creation
    and display of 3D visualizations based on the provided data files and
    offers various functionalities such as toggling display options and saving
    sessions.

    :param filenames: A list of file paths to be loaded. Supported file formats
        include .gev and .wrl. Only the first file is used if multiple .gev files
        are provided.
    :type filenames: list of str
    :param destination: The file path where the session will be saved. If not provided,
        the session is not saved. The file extension must be .gev if specified.
    :type destination: str, optional
    :param off_screen: If True, the plotter is created without displaying it. Defaults to False.
    :type off_screen: bool, optional
    :param safe_mode: If True, the viewer operates in safe mode with some features disabled.
    :type safe_mode: bool, optional
    :param no_warnings: If True, suppresses warning messages. Defaults to False.
    :type no_warnings: bool, optional

    :raises Exception: If .gev and .wrl files are mixed, or if multiple .gev files are provided.
    :raises Exception: If attempting to save a session using an invalid file extension.
    """

    def __init__(self, filenames, destination=None, off_screen=False,\
                 safe_mode=False, no_warnings=False):
        """Constructor method for the GeViewer object.
        """
        self.filenames = filenames
        self.safe_mode = safe_mode
        self.off_screen = off_screen
        self.no_warnings = no_warnings
        if not self.no_warnings:
            utils.check_for_updates()

        # if destination is given, the program will save the session to that file
        if destination is not None:
            self.save_session = True
            if not destination.endswith('.gev'):
                if self.off_screen:
                    print('Renaming the session file to end in .gev so it can be loaded later.')
                    destination = '.'.join(destination.split('.')[:-1]) + '.gev'
                else:
                    print('Error: invalid file extension.')
                    print('Try again, or press enter to continue without saving.\n')
                    destination = utils.prompt_for_file_path()
            if not os.path.isdir('/'.join(str(Path(destination).resolve()).split('/')[:-1])):
                if self.off_screen:
                    print('Destination folder does not exist.')
                    print('Creating the destination folder.')
                    os.makedirs('/'.join(str(Path(destination).resolve()).split('/')[:-1]), exist_ok=True)
                else:
                    print('Error: destination folder does not exist.')
                    print('Try again, or press enter to continue without saving.\n')
                    destination = utils.prompt_for_file_path()
            if destination is None:
                self.save_session = False
        else:
            self.save_session = False

        # ensure the input arguments are valid
        self.from_gev = True if filenames[0].endswith('.gev') else False
        if len(filenames) > 1:
            extensions = [Path(f).suffix for f in filenames]
            if not all([e == extensions[0] for e in extensions]):
                raise Exception('Cannot load .wrl and .gev files together.')
        if self.from_gev and len(filenames) > 1:
            print('Loading multiple .gev files at a time is not supported.')
            print('Only the first file will be loaded.\n')
            self.filenames = [self.filenames[0]]
        if self.from_gev and self.safe_mode:
            print('Safe mode can only be used for VRML files.')
            print('Ignoring the --safe-mode flag.\n')
            self.safe_mode = False
        if self.from_gev and self.save_session:
            print('This session has already been saved.')
            print('Ignoring the --destination flag.\n')
            self.save_session = False
        if destination is not None and self.safe_mode:
            print('Cannot save a session in safe mode.')
            print('Ignoring the --destination flag.\n')
            self.save_session = False

        if self.safe_mode:
            print('Running in safe mode with some features disabled.\n')
            self.view_params = (None, None, None)
            self.initial_camera_pos = None
            self.has_transparency = False
            self.create_plotter()
            if len(self.filenames)>1:
                print('Only the first file will be displayed in safe mode.\n')
            self.plotter.import_vrml(self.filenames[0])
            self.counts = []
            self.visible = []
            self.meshes = []
        else:
            self.visible = [True, True, True]
            if self.from_gev:
                self.load(filenames[0])
            else:
                data = utils.read_files(filenames)
                viewpoint_block, polyline_blocks, marker_blocks, solid_blocks = parser.extract_blocks(data)
                self.view_params = parser.parse_viewpoint_block(viewpoint_block)
                self.counts = [len(polyline_blocks), len(marker_blocks), len(solid_blocks)]
                if not self.save_session and not no_warnings and sum(self.counts) > 1e6:
                    self.save_session = utils.prompt_for_save_session(sum(self.counts))
                    if self.save_session:
                        destination = utils.prompt_for_file_path()
                self.meshes = parser.create_meshes(polyline_blocks, marker_blocks, solid_blocks)
            self.create_plotter()
            self.plot_meshes()
            self.set_initial_view()
            if self.save_session:
                self.save(destination)
        if not off_screen:
            self.show()

    
    def create_plotter(self):
        """Creates a Plotter object, a subclass of pyvista.Plotter.
        """
        self.plotter = plotter.Plotter(title='GeViewer -- ' + str(Path(self.filenames[0]).resolve()) \
                                       + ['',' + {} more'.format(len(self.filenames)-1)]\
                                         [(len(self.filenames)>1) and not self.safe_mode],\
                                       off_screen=self.off_screen)
        self.plotter.add_key_event('c', self.save_screenshot)
        self.plotter.add_key_event('t', self.toggle_tracks)
        self.plotter.add_key_event('m', self.toggle_step_markers)
        self.plotter.add_key_event('b', self.toggle_background)
        self.plotter.add_key_event('w', self.toggle_wireframe)
        self.plotter.add_key_event('d', self.set_window_size)
        self.plotter.add_key_event('i', self.set_camera_view)
        self.plotter.add_key_event('p', self.print_view_params)
        self.plotter.add_key_event('h', self.export_to_html)
        self.plotter.add_key_event('v', self.update_camera_position)
        self.plotter.set_background('lightskyblue',top='midnightblue')
        if not self.safe_mode:
            self.check_transparency()
        self.bkg_on = True
        self.wireframe = False


    def check_transparency(self):
        """Enables depth peeling if any of the meshes have transparency. This prevents issues
        with rendering order when displaying transparent objects.
        """
        transparencies = [mesh.point_data['color'][:,-1] for mesh in self.meshes if mesh is not None]
        transparencies = np.concatenate(transparencies)
        if not np.all(transparencies == 1):
            self.plotter.enable_depth_peeling()
            self.has_transparency = True
        else:
            self.has_transparency = False


    def plot_meshes(self):
        """Adds the meshes to the plot.
        """
        print('Plotting meshes...')
        actors = [None for i in range(3)]
        for i in range(3):
            if self.counts[i] > 0:
                actors[i] = self.plotter.add_mesh(self.meshes[i], scalars='color', rgba=True)
        self.actors = actors
        print('Done.\n')


    def set_initial_view(self):
        """Sets the initial camera viewpoint based on the view parameters
        provided in the VRML file.
        """
        fov = self.view_params[0]
        position = self.view_params[1]
        orientation = self.view_params[2]
        if position is not None and orientation is not None:
            up, v = utils.orientation_transform(orientation)
            focus = position + v*np.linalg.norm(position)
        elif position is not None and orientation is None:
            up = np.array([0.,1.,0.])
            focus = position + np.array([0.,0.,-1.])*np.linalg.norm(position)
        elif position is None and orientation is not None:
            self.plotter.view_isometric()
            position = np.linalg.norm(self.plotter.camera.GetPosition())*np.array([0.,0.,1.])
            up, v = utils.orientation_transform(orientation)
            focus = position + v*np.linalg.norm(position)
        else:
            self.plotter.view_isometric()
            position = np.linalg.norm(self.plotter.camera.GetPosition())*np.array([0.,0.,1.])
            up = np.array([0.,1.,0.])
            focus = position + np.array([0.,0.,-1.])*np.linalg.norm(position)
        self.plotter.reset_camera()
        self.set_camera_view((fov,position,up,focus))
        self.initial_camera_pos = self.plotter.camera_position
    

    def set_camera_view(self,args=None):
        """Sets the camera viewpoint.

        :param args: A list of the view parameters, defaults to None
        :type args: list, optional
        """
        if args is None:
            fov = None
            position, up, focus = asyncio.run(utils.prompt_for_camera_view())
        else:
            fov, position, up, focus = args
        if fov is not None:
            self.plotter.camera.view_angle = fov
        if position is not None:
            self.plotter.camera.position = position
        if up is not None:
            self.plotter.camera.up = up
        if focus is not None:
            self.plotter.camera.focal_point = focus
        if args is None:
            if not self.off_screen:
                self.plotter.update()
            print('Camera view set.\n')


    def update_camera_position(self):
        """Updates the camera position. This method's only job is to ensure
        that the PyVista plotter knows that the camera position has changed
        on a key event handled by vtk under the hood. This is necessary to
        avoid a bug where the camera jumps back to the isometric view the next
        time the user prints the camera position.
        """
        self.plotter.camera.GetPosition()


    def print_view_params(self):
        """Prints the current camera viewpoint parameters.
        """
        print('Viewpoint parameters:')
        print('  Window size: {}x{}'.format(*self.plotter.window_size))
        print('  Position:    ({}, {}, {})'.format(*self.plotter.camera.position))
        print('  Focal point: ({}, {}, {})'.format(*self.plotter.camera.focal_point))
        print('  Up vector:   ({}, {}, {})\n'.format(*self.plotter.camera.up))


    def save_screenshot(self):
        """Saves a screenshot of of the current view, prompting the user
        for a file path.
        """
        file_path = asyncio.run(utils.prompt_for_screenshot_path())
        if file_path is None:
            print('Operation cancelled.\n')
            return
        elif file_path.endswith('.png'):
            self.plotter.screenshot(file_path)
        else:
            self.plotter.save_graphic(file_path)
        print('Screenshot saved to ' + str(Path(file_path).resolve()) + '.\n')
    

    def set_window_size(self):
        """Sets the size of the viewer window in pixels, prompting the user
        for the width and height.
        """
        width, height = asyncio.run(utils.prompt_for_window_size())
        if width is None and height is None:
            print('Operation cancelled.\n')
            return
        self.plotter.window_size = width, height
        print('Window size set to ' + str(width) + 'x' + str(height) + '.\n')
        

    def toggle_tracks(self):
        """Toggles the particle trajectories on or off.
        """
        if not self.safe_mode:
            self.visible[0] = not self.visible[0]
            print('Toggling particle tracks ' + ['off.','on.'][self.visible[0]] + '\n')
            if self.visible[0]:
                if self.actors[0] is not None:
                    self.actors[0].visibility = True
            else:
                if self.actors[0] is not None:
                    self.actors[0].visibility = False
            if not self.off_screen:
                self.plotter.update()
        else:
            print('This feature is disabled in safe mode.\n')
                
                
    def toggle_step_markers(self):
        """Toggles the step markers on or off.
        """
        if not self.safe_mode:
            self.visible[1] = not self.visible[1]
            print('Toggling step markers ' + ['off.','on.'][self.visible[1]] + '\n')
            if self.visible[1]:
                if self.actors[1] is not None:
                    self.actors[1].visibility = True
            else:
                if self.actors[1] is not None:
                    self.actors[1].visibility = False
            if not self.off_screen:
                self.plotter.update()
        else:
            print('This feature is disabled in safe mode.\n')


    def toggle_background(self):
        """Toggles the gradient background on and off.
        """
        self.bkg_on = not self.bkg_on
        print('Toggling background ' + ['off.','on.'][self.bkg_on] + '\n')
        if self.bkg_on:
            self.plotter.set_background('lightskyblue',top='midnightblue')
        else:
            self.plotter.set_background('white')
        if not self.off_screen:
            self.plotter.update()


    def toggle_wireframe(self):
        """Toggles between solid and wireframe display modes. Disables depth
        peeling if wireframe mode is enabled to improve responsiveness.
        """
        self.wireframe = not self.wireframe
        print('Switching to ' + ['solid','wireframe'][self.wireframe] + ' mode.\n')
        if not self.safe_mode:
            if self.wireframe:
                self.actors[2].prop.SetRepresentationToWireframe()
                if self.has_transparency:
                    self.plotter.disable_depth_peeling()
            else:
                self.actors[2].prop.SetRepresentationToSurface()
                if self.has_transparency:
                    self.plotter.enable_depth_peeling()
        else:
            if self.wireframe:
                actors = self.plotter.renderer.GetActors()
                actors.InitTraversal()
                actor = actors.GetNextActor()
                while actor:
                    actor.GetProperty().SetRepresentationToWireframe()
                    actor = actors.GetNextActor()
            else:
                actors = self.plotter.renderer.GetActors()
                actors.InitTraversal()
                actor = actors.GetNextActor()
                while actor:
                    actor.GetProperty().SetRepresentationToSurface()
                    actor = actors.GetNextActor()
        if not self.off_screen:
            self.plotter.update()


    def export_to_html(self):
        """Saves the interactive viewer to an HTML file, prompting
        the user for a file path.
        """
        try:
            import nest_asyncio
            import trame
            import trame_vuetify
            import trame_vtk
        except ImportError:
            print('Error: exporting to HTML requires additional dependencies.')
            print('Run "pip install geviewer[extras]" to install them.\n')
            return
        file_path = asyncio.run(utils.prompt_for_html_path())
        if file_path is None:
            print('Operation cancelled.\n')
            return
        self.plotter.export_html(file_path)
        print('Interactive viewer saved to ' + str(Path(file_path).resolve()) + '.\n')


    def save(self, filename):
        """Saves the meshes to a .gev file.

        :param filename: The name of the file to save the session to.
        :type filename: str
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpfolder = tmpdir + '/gevfile/'
            os.makedirs(tmpfolder, exist_ok=False)
            for i,mesh in enumerate(self.meshes):
                if mesh is not None:
                    mesh.save(tmpfolder + 'mesh{}.vtk'.format(i))
            fov, pos, ori = self.view_params
            if fov is None:
                fov = 'None'
            if pos is None:
                pos = ['None' for i in range(3)]
            if ori is None:
                ori = ['None' for i in range(4)]
            viewpoint = np.array([str(fov)] + [str(p) for p in pos] + [str(o) for o in ori])
            np.save(tmpfolder + 'viewpoint.npy',np.array(viewpoint),allow_pickle=False)
            with zipfile.ZipFile(tmpdir + '/gevfile.gev', 'w') as archive:
                for file_name in os.listdir(tmpfolder):
                    file_path = os.path.join(tmpfolder, file_name)
                    archive.write(file_path, arcname=file_name)

            # if using the default filename and it exists, increment
            # the number until a unique filename is found
            if filename=='viewer.gev' and os.path.exists(filename):
                filename = 'viewer2.gev'
                i = 2
                while(os.path.exists('viewer{}.gev'.format(i))):
                    i += 1
                filename = 'viewer{}.gev'.format(i)
            shutil.copy(tmpdir + '/gevfile.gev', filename)
        print('Session saved to ' + str(Path(filename).resolve()) + '.\n')

                
    def load(self, filename):
        """Loads the meshes from a .gev file.

        :param filename: The path to the .gev file.
        :type filename: str
        :raises Exception: Invalid file format. Only .gev files are supported.
        """
        if not filename.endswith('.gev'):
            raise Exception('Invalid file format. Only .gev files are supported.')
        print('Loading session from ' + str(Path(filename).resolve()) + '...')
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpfolder = tmpdir + '/gevfile/'
            os.makedirs(tmpfolder, exist_ok=False)
            with zipfile.ZipFile(filename, 'r') as archive:
                archive.extractall(tmpfolder)
            meshes = []
            counts = []
            mesh_files = [file[-5] for file in os.listdir(tmpfolder) if file.endswith('.vtk')]
            for i in range(3):
                if str(i) not in mesh_files:
                    meshes.append(None)
                    counts.append(0)
                    continue
                meshes.append(pv.read(tmpfolder + 'mesh{}.vtk'.format(i)))
                counts.append(1)
            viewpoint = np.load(tmpfolder + 'viewpoint.npy')
            fov = float(viewpoint[0]) if viewpoint[0] != 'None' else None
            pos = [float(p) for p in viewpoint[1:4]] if viewpoint[1] != 'None' else None
            ori = [float(o) for o in viewpoint[4:]] if viewpoint[4] != 'None' else None
            self.view_params = (fov, pos, ori)
            self.meshes = meshes
            self.counts = counts


    def show(self):
        """Opens the plotting window.
        """
        self.plotter.show(cpos=self.initial_camera_pos,\
                          before_close_callback=lambda x: print('\nExiting GeViewer.\n'))



import numpy as np
import sys
import os
import ast
import asyncio
from pathlib import Path


def read_files(filenames):
    """Reads the content of multiple files and concatenates it into a single string.

    :param filenames: A list of file paths to read.
    :type filenames: list of str
    :return: A single string containing the concatenated content of all the files.
    :rtype: str
    """
    data = []
    for filename in filenames:
        print('Reading data from ' + str(Path(filename).resolve())+ '...')
        with open(filename, 'r') as f:
            for line in f:
                # don't read comments
                if not line.strip().startswith('#'):
                    data.append(line)
    data = ''.join(data)
    return data


def clear_input_buffer():
    """Clears the input buffer to prevent stray keystrokes from influencing
    subsequent inputs.

    :return: None
    """
    try:
        # if on Unix
        import termios
        termios.tcflush(sys.stdin, termios.TCIFLUSH)
    except ImportError:
        # if on Windows
        import msvcrt
        while msvcrt.kbhit():
            msvcrt.getch()


async def prompt_for_camera_view():
    """Asynchronously prompts the user for camera view input from the terminal.

    This function prompts the user to enter the camera's position, focal point, and
    up vector. Each input is expected to be a set of three comma-separated numbers.
    The function uses asynchronous input handling to avoid blocking other operations.

    :return: A tuple containing the position, up vector, and focal point. Each
        value is a list of three floating-point numbers or `None` if the user
        skips the input.
    :rtype: tuple
    """
    print('Setting the camera position and orientation.')
    print('Press enter to skip any of the following prompts.')
    clear_input_buffer()
    while(True):
        try:
            position = await asyncio.to_thread(input, 'Enter the position as three comma-separated numbers: ')
            if position == '':
                position = None
                break
            position = list(map(float, ast.literal_eval(position)))
            if len(position) != 3:
                raise ValueError
            break
        except ValueError:
            print('Error: invalid input. Please enter three numbers separated by commas.')
    while(True):
        try:
            focus = await asyncio.to_thread(input, 'Enter the focal point as three comma-separated numbers: ')
            if focus == '':
                focus = None
                break
            focus = list(map(float, ast.literal_eval(focus)))
            if len(focus) != 3:
                raise ValueError
            break
        except ValueError:
            print('Error: invalid input. Please enter three numbers separated by commas.')
    while(True):
        try:
            up = await asyncio.to_thread(input, 'Enter the up vector as three comma-separated numbers: ')
            if up == '':
                up = None
                break
            up = list(map(float, ast.literal_eval(up)))
            if len(up) != 3:
                raise ValueError
            break
        except ValueError:
            print('Error: invalid input. Please enter three numbers separated by commas.')
    return position, up, focus


async def prompt_for_screenshot_path():
    """Asynchronously prompts the user for a file path to save a screenshot.

    This function prompts the user to enter a file path where the screenshot will be saved.
    The accepted file formats are `.png`, `.svg`, `.eps`, `.ps`, `.pdf`, and `.tex`. The 
    function uses asynchronous input handling to avoid blocking other operations.

    :return: The file path as a string if a valid path is provided, or `None` if
        the input is canceled.
    :rtype: str or None
    """
    clear_input_buffer()
    print('Enter the destination file path to save the screenshot,')
    print('or press enter to cancel.')
    print('Accepted formats are .png, .svg, .eps, .ps, .pdf, and .tex.')
    while(True):
        try:
            file_path = await asyncio.to_thread(input,'Save as (e.g. /path/to/file.png): ')
            if file_path == '':
                return None
            if not file_path.endswith(('.png', '.svg', '.eps', '.ps', '.pdf', '.tex')):
                raise ValueError
            if not os.path.isdir('/'.join(str(Path(file_path).resolve()).split('/')[:-1])):
                raise ValueError
            break
        except ValueError:
            print('Error: invalid input. Please enter a valid file path.')
    return file_path


async def prompt_for_window_size():
    """Asynchronously prompts the user for the window size.

    This function prompts the user to enter the window size in pixels as two
    comma-separated integers, representing width and height. The function uses
    asynchronous input handling to avoid blocking other operations.

    :return: A tuple containing the width and height as integers if valid input is provided,
        or `(None, None)` if the input is canceled.
    :rtype: tuple of (int, int) or (None, None)
    """
    clear_input_buffer()
    while(True):
        try:
            print('Enter (width,height) in pixels as two comma-separated integers,')
            dims = await asyncio.to_thread(input, 'or press enter to cancel: ')
            if dims == '':
                return None, None
            dims = list(map(int, ast.literal_eval(dims)))
            width, height = dims
            return width, height
        except ValueError:
            print('Error: invalid input. Please enter an integer.')


def prompt_for_file_path():
    """Prompts the user for a file path to save the session.

    This function asks the user to enter a file path where the session should be saved.
    The file path should end with the `.gev` extension to ensure compatibility with
    GeViewer for later loading of the session.

    :return: The file path as a string if valid input is provided, or `None` if the
        input is canceled.
    :rtype: str or None
    """
    clear_input_buffer()
    print('Enter the destination file path to save the session,')
    print('or press enter to cancel. Use the file extension .gev')
    print('so GeViewer can load the session later.')
    while(True):
        try:
            file_path = input('Save as (e.g. /path/to/file.gev): ')
            if file_path == '':
                return None
            if not file_path.endswith('.gev'):
                raise ValueError
            if not os.path.isdir('/'.join(str(Path(file_path).resolve()).split('/')[:-1])):
                raise ValueError
            print()
            return file_path
        except ValueError:
            print('Error: invalid input. Please enter a valid file path')
            print('ending in .gev')


def prompt_for_save_session(total_meshes):
    """Prompts the user to decide whether to save the session before loading
    a large file.

    This function warns the user if the file they are attempting to view is large,
    based on the number of meshes (`total_meshes`). The user is given the option
    to save the current session to avoid reloading the file in the future.

    :param total_meshes: The number of meshes in the large file being loaded.
    :type total_meshes: int
    :return: A boolean indicating whether the session will be saved.
    :rtype: bool
    """
    clear_input_buffer()
    print('\nWarning: the file you are attempting to view is large')
    print('({} meshes) and may take a while to load. Do you want to save'.format(total_meshes))
    print('the session after loading to avoid having to reload it later?')
    while(True):
        try:
            save_input = input('Enter "y" or "n": ')
            if save_input.lower() not in ['y', 'n']:
                raise ValueError
            save_session = save_input.lower() == 'y'
            print('This session will ' + ['not ', ''][save_session] + 'be saved.')
            print('(To save the session by default, use the --destination flag.')
            print('To avoid this warning, use the --no-warnings flag.)\n')
            return save_session
        except ValueError:
            print('Error: invalid input. Please enter "y" or "n".')


async def prompt_for_html_path():
    """Asynchronously prompts the user to input a file path for saving an HTML file.

    This function prompts the user to enter a destination file path to save the viewer
    as an interactive HTML file.

    :return: The file path provided by the user if valid, or `None` if the user presses
        enter to cancel.
    :rtype: str or None
    """
    clear_input_buffer()
    print('Enter the destination file path to save the HTML file,')
    print('or press enter to cancel. Use the file extension .html.')
    while(True):
        try:
            file_path = await asyncio.to_thread(input, 'Save as (e.g. /path/to/file.html): ')
            if file_path == '':
                return None
            if not file_path.endswith('.html'):
                raise ValueError
            if not os.path.isdir('/'.join(str(Path(file_path).resolve()).split('/')[:-1])):
                raise ValueError
            return file_path
        except ValueError:
            print('Error: invalid input. Please enter a valid file path')
            print('ending in .html')


def orientation_transform(orientation):
    """Calculate the up and camera direction vectors based on the orientation.

    The function takes an orientation specified as a tuple or list `(x, y, z, theta)`,
    where `(x, y, z)` is the axis of rotation and `theta` is the angle of rotation in
    radians. It applies this rotation to the default up vector `(0, 1, 0)` and the
    default direction vector `(0, 0, -1)` to compute the new up and direction vectors.

    :param orientation: A tuple or list containing the axis of rotation and angle of rotation.
        The format is (x, y, z, theta), where (x, y, z) is the axis and theta is the rotation
        angle in radians.
    :type orientation: tuple or list of floats

    :return: A tuple containing the transformed up vector and the transformed direction vector.
    :rtype: tuple of (numpy.ndarray, numpy.ndarray)
    """
    v = orientation[:3]
    v = np.array(v)/np.linalg.norm(v)
    theta = orientation[3]
    up = np.array((v[0]*v[1]*(1 - np.cos(theta)) - v[2]*np.sin(theta),\
                   v[1]*v[1]*(1 - np.cos(theta)) + np.cos(theta),\
                   v[1]*v[2]*(1 - np.cos(theta)) + v[0]*np.sin(theta)))
    vprime = -np.array((v[0]*v[2]*(1 - np.cos(theta)) + v[1]*np.sin(theta),\
                        v[1]*v[2]*(1 - np.cos(theta)) - v[0]*np.sin(theta),\
                        v[2]*v[2]*(1 - np.cos(theta)) + np.cos(theta)))
    return up, vprime


def check_for_updates():
    """Determines whether the user is using the latest version of GeViewer.
    If not, prints a message to the console to inform the user.
    """
    try:
        import json
        from urllib import request
        import geviewer
        from packaging.version import parse
        url = 'https://pypi.python.org/pypi/geviewer/json'
        releases = json.loads(request.urlopen(url).read())['releases']
        versions = list(releases.keys())
        parsed = [parse(v) for v in versions]
        latest = parsed[parsed.index(max(parsed))]
        current = parse(geviewer.__version__)
        if current < latest and not (latest.is_prerelease or latest.is_postrelease or latest.is_devrelease):
            print('You are using GeViewer version {}. The latest version is {}.'.format(current, latest))
            print('Use "pip install --upgrade geviewer" to update to the latest version.\n')
    except:
        # don't want this to interrupt regular use if there's a problem
        return